@@ -1,5 +1,5 @@
 /*
- *   Copyright (C) 2009-2016 Erwin Waterlander
+ *   Copyright (C) 2009-2017 Erwin Waterlander
  *   All rights reserved.
  *
  *   Redistribution and use in source and binary forms, with or without
@@ -56,7 +56,42 @@
 #ifdef D2U_UNIFILE
 int d2u_display_encoding = D2U_DISPLAY_ANSI ;
 #endif
-int header_done = 0;
+
+/* Copy string src to dest, and null terminate dest.
+   dest_size must be the buffer size of dest. */
+char *d2u_strncpy(char *dest, const char *src, size_t dest_size)
+{
+    strncpy(dest,src,dest_size);
+    dest[dest_size-1] = '\0';
+#ifdef DEBUG
+    if(strlen(src) > (dest_size-1)) {
+        D2U_UTF8_FPRINTF(stderr, "Text %s has been truncated from %d to %d characters in %s to prevent a buffer overflow.\n", src, (int)strlen(src), (int)dest_size, "d2u_strncpy()");
+    }
+#endif
+    return dest;
+}
+
+int d2u_fclose (FILE *fp, const char *filename, CFlag *ipFlag, const char *m, const char *progname)
+{
+  if (fclose(fp) != 0) {
+    if (ipFlag->verbose) {
+      ipFlag->error = errno;
+      D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
+      if (m[0] == 'w')
+        D2U_UTF8_FPRINTF(stderr, _("Failed to write to temporary output file %s:"), filename);
+      else
+        D2U_UTF8_FPRINTF(stderr, _("Failed to close input file %s:"), filename);
+      D2U_ANSI_FPRINTF(stderr, " %s\n", strerror(errno));
+    }
+    return EOF;
+  }
+#if DEBUG
+  else
+     fprintf(stderr, "%s: Closing file \"%s\" OK.\n", progname, filename);
+#endif
+  return 0;
+}
+
 
 /*
  * Print last system error on Windows.
@@ -609,6 +644,9 @@ int is_dos2unix(const char *progname)
 void PrintUsage(const char *progname)
 {
   D2U_ANSI_FPRINTF(stdout,_("Usage: %s [options] [file ...] [-n infile outfile ...]\n"), progname);
+#ifndef NO_CHOWN
+  D2U_ANSI_FPRINTF(stdout,_(" --allow-chown         allow file ownership change\n"));
+#endif
   D2U_ANSI_FPRINTF(stdout,_(" -ascii                convert only line breaks (default)\n"));
   D2U_ANSI_FPRINTF(stdout,_(" -iso                  conversion between DOS and ISO-8859-1 character set\n"));
   D2U_ANSI_FPRINTF(stdout,_("   -1252               use Windows code page 1252 (Western European)\n"));
@@ -644,6 +682,9 @@ void PrintUsage(const char *progname)
   D2U_ANSI_FPRINTF(stdout,_(" -n, --newfile         write to new file\n\
    infile              original file in new-file mode\n\
    outfile             output file in new-file mode\n"));
+#ifndef NO_CHOWN
+  D2U_ANSI_FPRINTF(stdout,_(" --no-allow-chown      don't allow file ownership change (default)\n"));
+#endif
   D2U_ANSI_FPRINTF(stdout,_(" -o, --oldfile         write to old file (default)\n\
    file ...            files to convert in old-file mode\n"));
   D2U_ANSI_FPRINTF(stdout,_(" -q, --quiet           quiet mode, suppress all warnings\n"));
@@ -728,6 +769,11 @@ void PrintVersion(const char *progname, const char *localedir)
 #else
   D2U_ANSI_FPRINTF(stdout,"%s", "Without native language support.\n");
 #endif
+#ifndef NO_CHOWN
+  D2U_ANSI_FPRINTF(stdout,"%s", _("With support to preserve the user and group ownership of files.\n"));
+#else
+  D2U_ANSI_FPRINTF(stdout,"%s", _("Without support to preserve the user and group ownership of files.\n"));
+#endif
 #ifdef ENABLE_NLS
   D2U_ANSI_FPRINTF(stdout,"LOCALEDIR: %s\n", localedir);
 #endif
@@ -735,8 +781,8 @@ void PrintVersion(const char *progname, const char *localedir)
 }
 
 /* opens file of name ipFN in read only mode
- * RetVal: NULL if failure
- *         file stream otherwise
+ * returns: NULL if failure
+ *          file stream otherwise
  */
 FILE* OpenInFile(char *ipFN)
 {
@@ -752,8 +798,8 @@ FILE* OpenInFile(char *ipFN)
 
 
 /* opens file of name opFN in write only mode
- * RetVal: NULL if failure
- *         file stream otherwise
+ * returns: NULL if failure
+ *          file stream otherwise
  */
 FILE* OpenOutFile(char *opFN)
 {
@@ -768,8 +814,8 @@ FILE* OpenOutFile(char *opFN)
 }
 
 /* opens file descriptor in write only mode
- * RetVal: NULL if failure
- *         file stream otherwise
+ * returns: NULL if failure
+ *          file stream otherwise
  */
 FILE* OpenOutFiled(int fd)
 {
@@ -800,6 +846,7 @@ char *dirname(char *path)
   return path;
 }
 
+#ifdef NO_MKSTEMP
 char *basename(char *path)
 {
   char *ptr;
@@ -821,6 +868,7 @@ char *basename(char *path)
    return ptr ;
 }
 #endif
+#endif
 
 /* Standard mktemp() is not safe to use (See mktemp(3)).
  * On Windows it is recommended to use GetTempFileName() (See MSDN).
@@ -893,7 +941,7 @@ char *d2u_mktemp(char *template)
   fname_str = (char *)malloc(len);
   if (! fname_str)
     return NULL;
-  strcpy(fname_str, szTempFileName);
+  d2u_strncpy(fname_str, szTempFileName,len);
 #endif
   /* replace all back slashes with slashes */
   while ( (ptr = strchr(fname_str,'\\')) != NULL)
@@ -966,8 +1014,8 @@ FILE* MakeTempFileFrom(const char *OutFN, char **fname_ret)
  * Note that if symbolic links are not supported, then 0 is always returned
  * and *rFN = lFN.
  *
- * RetVal: 0 if success, and *lFN is not a symlink
- *         1 if success, and *lFN is a symlink
+ * returns: 0 if success, and *lFN is not a symlink
+ *          1 if success, and *lFN is a symlink
  *         -1 otherwise
  */
 int ResolveSymbolicLink(char *lFN, char **rFN, CFlag *ipFlag, const char *progname)
@@ -1192,16 +1240,16 @@ void print_bom (const int bomtype, const char *filename, const char *progname)
 
     switch (bomtype) {
     case FILE_UTF16LE:   /* UTF-16 Little Endian */
-      strncpy(informat,_("UTF-16LE"),sizeof(informat));
+      d2u_strncpy(informat,_("UTF-16LE"),sizeof(informat));
       break;
     case FILE_UTF16BE:   /* UTF-16 Big Endian */
-      strncpy(informat,_("UTF-16BE"),sizeof(informat));
+      d2u_strncpy(informat,_("UTF-16BE"),sizeof(informat));
       break;
     case FILE_UTF8:      /* UTF-8 */
-      strncpy(informat,_("UTF-8"),sizeof(informat));
+      d2u_strncpy(informat,_("UTF-8"),sizeof(informat));
       break;
     case FILE_GB18030:      /* GB18030 */
-      strncpy(informat,_("GB18030"),sizeof(informat));
+      d2u_strncpy(informat,_("GB18030"),sizeof(informat));
       break;
     default:
     ;
@@ -1252,10 +1300,15 @@ void print_bom_info (const int bomtype)
   }
 }
 
+/* check_unicode_info()
+ * Print assumed encoding and read file's BOM. Return file's BOM in *bomtype_orig.
+ * Set ipFlag->bomtype to assumed BOM type, when file's BOM == FILE_MBS.
+ * Return -1 when a read error occurred, or when whar_t < 32 bit on non-Windows OS.
+ * Return 0 when everything is OK.
+ */
+
 int check_unicode_info(FILE *InF, CFlag *ipFlag, const char *progname, int *bomtype_orig)
 {
-  int RetVal = 0;
-
 #ifdef D2U_UNICODE
   if (ipFlag->verbose > 1) {
     if (ipFlag->ConvMode == CONVMODE_UTF16LE) {
@@ -1286,18 +1339,17 @@ int check_unicode_info(FILE *InF, CFlag *ipFlag, const char *progname, int *bomt
       /* A decoded UTF-16 surrogate pair must fit in a wchar_t */
       ipFlag->status |= WCHAR_T_TOO_SMALL ;
       if (!ipFlag->error) ipFlag->error = 1;
-      RetVal = -1;
+      return -1;
     }
   }
 #endif
 #endif
 
-  return RetVal;
+  return 0;
 }
 
 int check_unicode(FILE *InF, FILE *TempF,  CFlag *ipFlag, const char *ipInFN, const char *progname)
 {
-  int RetVal = 0;
 
 #ifdef D2U_UNICODE
   if (ipFlag->verbose > 1) {
@@ -1340,7 +1392,7 @@ int check_unicode(FILE *InF, FILE *TempF,  CFlag *ipFlag, const char *ipInFN, co
       /* A decoded UTF-16 surrogate pair must fit in a wchar_t */
       ipFlag->status |= WCHAR_T_TOO_SMALL ;
       if (!ipFlag->error) ipFlag->error = 1;
-      RetVal = -1;
+      return -1;
     }
   }
 #endif
@@ -1351,14 +1403,14 @@ int check_unicode(FILE *InF, FILE *TempF,  CFlag *ipFlag, const char *ipInFN, co
 #endif
 #endif
 
-  if ((!RetVal) && ((ipFlag->add_bom) || ((ipFlag->keep_bom) && (ipFlag->bomtype > 0))))
+  if ((ipFlag->add_bom) || ((ipFlag->keep_bom) && (ipFlag->bomtype > 0)))
     if (write_bom(TempF, ipFlag, progname) == NULL) return -1;
 
-  return RetVal;
+  return 0;
 }
 
 /* convert file ipInFN and write to file ipOutFN
- * RetVal: 0 if success
+ * returns: 0 if success
  *         -1 otherwise
  */
 int ConvertNewFile(char *ipInFN, char *ipOutFN, CFlag *ipFlag, const char *progname,
@@ -1472,7 +1524,7 @@ int ConvertNewFile(char *ipInFN, char *ipOutFN, CFlag *ipFlag, const char *progn
     if (check_unicode(InF, TempF, ipFlag, ipInFN, progname))
       RetVal = -1;
 
-  /* conversion sucessful? */
+  /* conversion successful? */
 #ifdef D2U_UNICODE
   if ((ipFlag->bomtype == FILE_UTF16LE) || (ipFlag->bomtype == FILE_UTF16BE)) {
     if ((!RetVal) && (ConvertW(InF, TempF, ipFlag, progname)))
@@ -1491,29 +1543,13 @@ int ConvertNewFile(char *ipInFN, char *ipOutFN, CFlag *ipFlag, const char *progn
 #endif
 
    /* can close in file? */
-  if (fclose(InF) == EOF) {
-    if (ipFlag->verbose) {
-      ipFlag->error = errno;
-      errstr = strerror(errno);
-      D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
-      D2U_UTF8_FPRINTF(stderr, _("Failed to close input file %s:"), ipInFN);
-      D2U_ANSI_FPRINTF(stderr, " %s\n", errstr);
-    }
+  if (d2u_fclose(InF, ipInFN, ipFlag, "r", progname) == EOF)
     RetVal = -1;
-  }
 
   /* can close output file? */
   if (TempF) {
-    if (fclose(TempF) == EOF) {
-      if (ipFlag->verbose) {
-        ipFlag->error = errno;
-        errstr = strerror(errno);
-        D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
-        D2U_UTF8_FPRINTF(stderr, _("Failed to write to temporary output file %s:"), TempPath);
-        D2U_ANSI_FPRINTF(stderr, " %s\n", errstr);
-      }
+    if (d2u_fclose(TempF, TempPath, ipFlag, "w", progname) == EOF)
       RetVal = -1;
-    }
   }
 
 #ifndef NO_CHMOD
@@ -1545,14 +1581,36 @@ int ConvertNewFile(char *ipInFN, char *ipOutFN, CFlag *ipFlag, const char *progn
      /* Required when a different user (e.g. root) has write permission on the original file. */
      /* Make sure that the original owner can still access the file. */
      if (chown(TempPath, StatBuf.st_uid, StatBuf.st_gid)) {
-        if (ipFlag->verbose) {
-          ipFlag->error = errno;
-          errstr = strerror(errno);
-          D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
-          D2U_UTF8_FPRINTF(stderr, _("Failed to change the owner and group of temporary output file %s:"), TempPath);
-          D2U_ANSI_FPRINTF(stderr, " %s\n", errstr);
+        if (ipFlag->AllowChown) {
+          if (ipFlag->verbose) {
+            D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
+            D2U_UTF8_FPRINTF(stderr, _("The user and/or group ownership of file %s is not preserved.\n"), ipOutFN);
+          }
+#ifndef NO_CHMOD
+          /* Set read/write permissions same as in new file mode. */
+          mask = umask(0); /* get process's umask */
+          umask(mask); /* set umask back to original */
+          RetVal = chmod(TempPath, StatBuf.st_mode & ~mask); /* set original permissions, minus umask */
+          if (RetVal) {
+             if (ipFlag->verbose) {
+               ipFlag->error = errno;
+               errstr = strerror(errno);
+               D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
+               D2U_UTF8_FPRINTF(stderr, _("Failed to change the permissions of temporary output file %s:"), TempPath);
+               D2U_ANSI_FPRINTF(stderr, " %s\n", errstr);
+             }
+          }
+#endif
+        } else {
+          if (ipFlag->verbose) {
+            ipFlag->error = errno;
+            errstr = strerror(errno);
+            D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
+            D2U_UTF8_FPRINTF(stderr, _("Failed to change the owner and group of temporary output file %s:"), TempPath);
+            D2U_ANSI_FPRINTF(stderr, " %s\n", errstr);
+          }
+          RetVal = -1;
         }
-        RetVal = -1;
      }
   }
 #endif
@@ -1642,7 +1700,7 @@ int ConvertNewFile(char *ipInFN, char *ipOutFN, CFlag *ipFlag, const char *progn
 }
 
 /* convert stdin and write to stdout
- * RetVal: 0 if success
+ * returns: 0 if success
  *         -1 otherwise
  */
 int ConvertStdio(CFlag *ipFlag, const char *progname,
@@ -1715,37 +1773,37 @@ void print_format(const CFlag *pFlag, char *informat, char *outformat, size_t li
   outformat[0]='\0';
 
   if (pFlag->bomtype == FILE_UTF16LE)
-    strncpy(informat,_("UTF-16LE"),lin);
+    d2u_strncpy(informat,_("UTF-16LE"),lin);
   if (pFlag->bomtype == FILE_UTF16BE)
-    strncpy(informat,_("UTF-16BE"),lin);
+    d2u_strncpy(informat,_("UTF-16BE"),lin);
   informat[lin-1]='\0';
 
 #ifdef D2U_UNICODE
   if ((pFlag->bomtype == FILE_UTF16LE)||(pFlag->bomtype == FILE_UTF16BE)) {
 #if !defined(__MSDOS__) && !defined(_WIN32) && !defined(__OS2__)  /* Unix, Cygwin */
-    strncpy(outformat,nl_langinfo(CODESET),lout);
+    d2u_strncpy(outformat,nl_langinfo(CODESET),lout);
 #endif
 
 #if defined(_WIN32) && !defined(__CYGWIN__) /* Windows, not Cygwin */
     if (pFlag->locale_target == TARGET_GB18030)
-      strncpy(outformat, _("GB18030"),lout);
+      d2u_strncpy(outformat, _("GB18030"),lout);
     else
-      strncpy(outformat, _("UTF-8"),lout);
+      d2u_strncpy(outformat, _("UTF-8"),lout);
 #endif
 
     if (pFlag->keep_utf16)
     {
       if (pFlag->bomtype == FILE_UTF16LE)
-        strncpy(outformat,_("UTF-16LE"),lout);
+        d2u_strncpy(outformat,_("UTF-16LE"),lout);
       if (pFlag->bomtype == FILE_UTF16BE)
-        strncpy(outformat,_("UTF-16BE"),lout);
+        d2u_strncpy(outformat,_("UTF-16BE"),lout);
     }
     outformat[lout-1]='\0';
   }
 #endif
 }
 
-void print_messages_newfile(const CFlag *pFlag, const char *infile, const char *outfile, const char *progname, const int RetVal)
+void print_messages(const CFlag *pFlag, const char *infile, const char *outfile, const char *progname, const int conversion_error)
 {
   char informat[32];
   char outformat[64];
@@ -1773,11 +1831,14 @@ void print_messages_newfile(const CFlag *pFlag, const char *infile, const char *
     D2U_UTF8_FPRINTF(stderr, _("Skipping %s, not a regular file.\n"), infile);
   } else if (pFlag->status & OUTPUTFILE_SYMLINK) {
     D2U_UTF8_FPRINTF(stderr,"%s: ",progname);
-    D2U_UTF8_FPRINTF(stderr, _("Skipping %s, output file %s is a symbolic link.\n"), infile, outfile);
+    if (outfile)
+      D2U_UTF8_FPRINTF(stderr, _("Skipping %s, output file %s is a symbolic link.\n"), infile, outfile);
+    else
+      D2U_UTF8_FPRINTF(stderr, _("Skipping symbolic link %s.\n"), infile);
   } else if (pFlag->status & INPUT_TARGET_NO_REGFILE) {
     D2U_UTF8_FPRINTF(stderr,"%s: ",progname);
     D2U_UTF8_FPRINTF(stderr, _("Skipping symbolic link %s, target is not a regular file.\n"), infile);
-  } else if (pFlag->status & OUTPUT_TARGET_NO_REGFILE) {
+  } else if ((pFlag->status & OUTPUT_TARGET_NO_REGFILE) && outfile) {
     D2U_UTF8_FPRINTF(stderr,"%s: ",progname);
     D2U_UTF8_FPRINTF(stderr, _("Skipping %s, target of symbolic link %s is not a regular file.\n"), infile, outfile);
   } else if (pFlag->status & BINARY_FILE) {
@@ -1799,122 +1860,64 @@ void print_messages_newfile(const CFlag *pFlag, const char *infile, const char *
     D2U_UTF8_FPRINTF(stderr, _("Skipping UTF-16 file %s, UTF-16 conversion is not supported in this version of %s.\n"), infile, progname);
 #endif
   } else {
-    D2U_UTF8_FPRINTF(stderr,"%s: ",progname);
-    if (informat[0] == '\0') {
-      if (is_dos2unix(progname)) {
-        if (!RetVal) D2U_UTF8_FPRINTF(stderr, _("converting file %s to file %s in Unix format...\n"), infile, outfile);
-      } else {
-        if (pFlag->FromToMode == FROMTO_UNIX2MAC) {
-          if (!RetVal) D2U_UTF8_FPRINTF(stderr, _("converting file %s to file %s in Mac format...\n"), infile, outfile);
+    if (!conversion_error) {
+      D2U_UTF8_FPRINTF(stderr,"%s: ",progname);
+      if (informat[0] == '\0') {
+        if (is_dos2unix(progname)) {
+          if (outfile)
+            D2U_UTF8_FPRINTF(stderr, _("converting file %s to file %s in Unix format...\n"), infile, outfile);
+          else
+            D2U_UTF8_FPRINTF(stderr, _("converting file %s to Unix format...\n"), infile);
         } else {
-          if (!RetVal) D2U_UTF8_FPRINTF(stderr, _("converting file %s to file %s in DOS format...\n"), infile, outfile);
+          if (pFlag->FromToMode == FROMTO_UNIX2MAC) {
+            if (outfile)
+              D2U_UTF8_FPRINTF(stderr, _("converting file %s to file %s in Mac format...\n"), infile, outfile);
+            else
+              D2U_UTF8_FPRINTF(stderr, _("converting file %s to Mac format...\n"), infile);
+          } else {
+            if (outfile)
+              D2U_UTF8_FPRINTF(stderr, _("converting file %s to file %s in DOS format...\n"), infile, outfile);
+            else
+              D2U_UTF8_FPRINTF(stderr, _("converting file %s to DOS format...\n"), infile);
+          }
         }
-      }
-    } else {
-      if (is_dos2unix(progname)) {
+      } else {
+        if (is_dos2unix(progname)) {
+          if (outfile)
     /* TRANSLATORS:
 1st %s is encoding of input file.
 2nd %s is name of input file.
 3rd %s is encoding of output file.
 4th %s is name of output file.
 E.g.: converting UTF-16LE file in.txt to UTF-8 file out.txt in Unix format... */
-        if (!RetVal) D2U_UTF8_FPRINTF(stderr, _("converting %s file %s to %s file %s in Unix format...\n"), informat, infile, outformat, outfile);
-      } else {
-        if (pFlag->FromToMode == FROMTO_UNIX2MAC) {
-          if (!RetVal) D2U_UTF8_FPRINTF(stderr, _("converting %s file %s to %s file %s in Mac format...\n"), informat, infile, outformat, outfile);
-        } else {
-          if (!RetVal) D2U_UTF8_FPRINTF(stderr, _("converting %s file %s to %s file %s in DOS format...\n"), informat, infile, outformat, outfile);
-        }
-      }
-    }
-    if (RetVal) {
-      D2U_UTF8_FPRINTF(stderr,"%s: ",progname);
-      D2U_UTF8_FPRINTF(stderr, _("problems converting file %s to file %s\n"), infile, outfile);
-    }
-  }
-}
-
-void print_messages_oldfile(const CFlag *pFlag, const char *infile, const char *progname, const int RetVal)
-{
-  char informat[10];
-  char outformat[32];
-# ifdef D2U_UNIFILE
-  wchar_t informatw[32];
-  wchar_t outformatw[64];
-#endif
-
-  print_format(pFlag, informat, outformat, sizeof(informat), sizeof(outformat));
-
-/* Change informat and outformat to UTF-8 for d2u_utf8_fprintf. */
-# ifdef D2U_UNIFILE
-   /* The format string is encoded in the system default
-    * Windows ANSI code page. May have been translated
-    * by gettext. Convert it to wide characters. */
-   d2u_MultiByteToWideChar(CP_ACP,0, informat, -1, informatw, sizeof(informat));
-   d2u_MultiByteToWideChar(CP_ACP,0, outformat, -1, outformatw, sizeof(outformat));
-   /* then convert the format string to UTF-8 */
-   d2u_WideCharToMultiByte(CP_UTF8, 0, informatw, -1, informat, sizeof(informat), NULL, NULL);
-   d2u_WideCharToMultiByte(CP_UTF8, 0, outformatw, -1, outformat, sizeof(outformat), NULL, NULL);
-#endif
-
-  if (pFlag->status & NO_REGFILE) {
-    D2U_UTF8_FPRINTF(stderr,"%s: ",progname);
-    D2U_UTF8_FPRINTF(stderr, _("Skipping %s, not a regular file.\n"), infile);
-  } else if (pFlag->status & OUTPUTFILE_SYMLINK) {
-    D2U_UTF8_FPRINTF(stderr,"%s: ",progname);
-    D2U_UTF8_FPRINTF(stderr, _("Skipping symbolic link %s.\n"), infile);
-  } else if (pFlag->status & INPUT_TARGET_NO_REGFILE) {
-    D2U_UTF8_FPRINTF(stderr,"%s: ",progname);
-    D2U_UTF8_FPRINTF(stderr, _("Skipping symbolic link %s, target is not a regular file.\n"), infile);
-  } else if (pFlag->status & BINARY_FILE) {
-    D2U_UTF8_FPRINTF(stderr,"%s: ",progname);
-    D2U_UTF8_FPRINTF(stderr, _("Skipping binary file %s\n"), infile);
-  } else if (pFlag->status & WRONG_CODEPAGE) {
-    D2U_UTF8_FPRINTF(stderr,"%s: ",progname);
-    D2U_UTF8_FPRINTF(stderr, _("code page %d is not supported.\n"), pFlag->ConvMode);
-#ifdef D2U_UNICODE
-  } else if (pFlag->status & WCHAR_T_TOO_SMALL) {
-    D2U_UTF8_FPRINTF(stderr,"%s: ",progname);
-    D2U_UTF8_FPRINTF(stderr, _("Skipping UTF-16 file %s, the size of wchar_t is %d bytes.\n"), infile, (int)sizeof(wchar_t));
-  } else if (pFlag->status & UNICODE_CONVERSION_ERROR) {
-    D2U_UTF8_FPRINTF(stderr,"%s: ",progname);
-    D2U_UTF8_FPRINTF(stderr, _("Skipping UTF-16 file %s, an UTF-16 conversion error occurred on line %u.\n"), infile, pFlag->line_nr);
-#else
-  } else if (pFlag->status & UNICODE_NOT_SUPPORTED) {
-    D2U_UTF8_FPRINTF(stderr,"%s: ",progname);
-    D2U_UTF8_FPRINTF(stderr, _("Skipping UTF-16 file %s, UTF-16 conversion is not supported in this version of %s.\n"), infile, progname);
-#endif
-  } else {
-    D2U_UTF8_FPRINTF(stderr,"%s: ",progname);
-    if (informat[0] == '\0') {
-      if (is_dos2unix(progname)) {
-        if (!RetVal) D2U_UTF8_FPRINTF(stderr, _("converting file %s to Unix format...\n"), infile);
-      } else {
-        if (pFlag->FromToMode == FROMTO_UNIX2MAC) {
-          if (!RetVal) D2U_UTF8_FPRINTF(stderr, _("converting file %s to Mac format...\n"), infile);
-        } else {
-          if (!RetVal) D2U_UTF8_FPRINTF(stderr, _("converting file %s to DOS format...\n"), infile);
-        }
-      }
-    } else {
-      if (is_dos2unix(progname)) {
+            D2U_UTF8_FPRINTF(stderr, _("converting %s file %s to %s file %s in Unix format...\n"), informat, infile, outformat, outfile);
+          else
     /* TRANSLATORS:
 1st %s is encoding of input file.
 2nd %s is name of input file.
 3rd %s is encoding of output (input file is overwritten).
 E.g.: converting UTF-16LE file foo.txt to UTF-8 Unix format... */
-        if (!RetVal) D2U_UTF8_FPRINTF(stderr, _("converting %s file %s to %s Unix format...\n"), informat, infile, outformat);
-      } else {
-        if (pFlag->FromToMode == FROMTO_UNIX2MAC) {
-          if (!RetVal) D2U_UTF8_FPRINTF(stderr, _("converting %s file %s to %s Mac format...\n"), informat, infile, outformat);
+            D2U_UTF8_FPRINTF(stderr, _("converting %s file %s to %s Unix format...\n"), informat, infile, outformat);
         } else {
-          if (!RetVal) D2U_UTF8_FPRINTF(stderr, _("converting %s file %s to %s DOS format...\n"), informat, infile, outformat);
+          if (pFlag->FromToMode == FROMTO_UNIX2MAC) {
+            if (outfile)
+              D2U_UTF8_FPRINTF(stderr, _("converting %s file %s to %s file %s in Mac format...\n"), informat, infile, outformat, outfile);
+            else
+              D2U_UTF8_FPRINTF(stderr, _("converting %s file %s to %s Mac format...\n"), informat, infile, outformat);
+          } else {
+            if (outfile)
+              D2U_UTF8_FPRINTF(stderr, _("converting %s file %s to %s file %s in DOS format...\n"), informat, infile, outformat, outfile);
+            else
+              D2U_UTF8_FPRINTF(stderr, _("converting %s file %s to %s DOS format...\n"), informat, infile, outformat);
+          }
         }
       }
-    }
-    if (RetVal) {
+    } else {
       D2U_UTF8_FPRINTF(stderr,"%s: ",progname);
-      D2U_UTF8_FPRINTF(stderr, _("problems converting file %s\n"), infile);
+      if (outfile)
+        D2U_UTF8_FPRINTF(stderr, _("problems converting file %s to file %s\n"), infile, outfile);
+      else
+        D2U_UTF8_FPRINTF(stderr, _("problems converting file %s\n"), infile);
     }
   }
 }
@@ -1943,6 +1946,8 @@ void print_messages_info(const CFlag *pFlag, const char *infile, const char *pro
 
 void printInfo(CFlag *ipFlag, const char *filename, int bomtype, unsigned int lb_dos, unsigned int lb_unix, unsigned int lb_mac)
 {
+  static int header_done = 0;
+
   if (ipFlag->file_info & INFO_CONVERT) {
     if ((ipFlag->FromToMode == FROMTO_DOS2UNIX) && (lb_dos == 0))
       return;
@@ -1965,9 +1970,15 @@ void printInfo(CFlag *ipFlag, const char *filename, int bomtype, unsigned int lb
       D2U_UTF8_FPRINTF(stdout, "  BOM     ");
     if (ipFlag->file_info & INFO_TEXT)
       D2U_UTF8_FPRINTF(stdout, "  TXTBIN");
-    if (*filename != '\0')
-      D2U_UTF8_FPRINTF(stdout, "  FILE");
-    D2U_UTF8_FPRINTF(stdout, "\n");
+    if (*filename != '\0') {
+      if (ipFlag->file_info & INFO_DEFAULT)
+        D2U_UTF8_FPRINTF(stdout, "  ");
+      D2U_UTF8_FPRINTF(stdout, "FILE");
+    }
+    if (ipFlag->file_info & INFO_PRINT0)
+      fputc(0, stdout);
+    else
+      D2U_UTF8_FPRINTF(stdout, "\n");
     header_done = 1;
   }
 
@@ -1991,9 +2002,14 @@ void printInfo(CFlag *ipFlag, const char *filename, int bomtype, unsigned int lb
       ptr++;
     else
       ptr = filename;
-    D2U_UTF8_FPRINTF(stdout, "  %s",ptr);
+    if (ipFlag->file_info & INFO_DEFAULT)
+      D2U_UTF8_FPRINTF(stdout, "  ");
+    D2U_UTF8_FPRINTF(stdout, "%s",ptr);
   }
-  D2U_UTF8_FPRINTF(stdout, "\n");
+  if (ipFlag->file_info & INFO_PRINT0)
+    fputc(0, stdout);
+  else
+    D2U_UTF8_FPRINTF(stdout, "\n");
 }
 
 #ifdef D2U_UNICODE
@@ -2097,7 +2113,6 @@ void FileInfo(FILE* ipInF, CFlag *ipFlag, const char *filename, int bomtype, con
 
 int GetFileInfo(char *ipInFN, CFlag *ipFlag, const char *progname)
 {
-  int RetVal = 0;
   FILE *InF = NULL;
   int bomtype_orig = FILE_MBS; /* messages must print the real bomtype, not the assumed bomtype */
 
@@ -2127,46 +2142,35 @@ int GetFileInfo(char *ipInFN, CFlag *ipFlag, const char *progname)
       D2U_UTF8_FPRINTF(stderr, "%s: %s: ", progname, ipInFN);
       D2U_ANSI_FPRINTF(stderr, "%s\n", errstr);
     }
-    RetVal = -1;
+    return -1;
   }
 
 
-  if (!RetVal)
-    if (check_unicode_info(InF, ipFlag, progname, &bomtype_orig))
-      RetVal = -1;
+  if (check_unicode_info(InF, ipFlag, progname, &bomtype_orig)) {
+    d2u_fclose(InF, ipInFN, ipFlag, "r", progname);
+    return -1;
+  }
 
-  /* info sucessful? */
+  /* info successful? */
 #ifdef D2U_UNICODE
-  if (!RetVal) {
-    if ((ipFlag->bomtype == FILE_UTF16LE) || (ipFlag->bomtype == FILE_UTF16BE)) {
-      FileInfoW(InF, ipFlag, ipInFN, bomtype_orig, progname);
-    } else {
-      FileInfo(InF, ipFlag, ipInFN, bomtype_orig, progname);
-    }
+  if ((ipFlag->bomtype == FILE_UTF16LE) || (ipFlag->bomtype == FILE_UTF16BE)) {
+    FileInfoW(InF, ipFlag, ipInFN, bomtype_orig, progname);
+  } else {
+    FileInfo(InF, ipFlag, ipInFN, bomtype_orig, progname);
   }
 #else
-  if (!RetVal)
-    FileInfo(InF, ipFlag, ipInFN, bomtype_orig, progname);
+  FileInfo(InF, ipFlag, ipInFN, bomtype_orig, progname);
 #endif
 
   /* can close in file? */
-  if (fclose(InF) == EOF) {
-    if (ipFlag->verbose) {
-      char *errstr = strerror(errno);
-      ipFlag->error = errno;
-      D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
-      D2U_UTF8_FPRINTF(stderr, _("Failed to close input file %s:"), ipInFN);
-      D2U_ANSI_FPRINTF(stderr, " %s\n", errstr);
-    }
-    RetVal = -1;
-  }
+  if (d2u_fclose(InF, ipInFN, ipFlag, "r", progname) == EOF)
+    return -1;
 
-  return RetVal;
+  return 0;
 }
 
 int GetFileInfoStdio(CFlag *ipFlag, const char *progname)
 {
-  int RetVal = 0;
   int bomtype_orig = FILE_MBS; /* messages must print the real bomtype, not the assumed bomtype */
 
   ipFlag->status = 0 ;
@@ -2187,23 +2191,20 @@ int GetFileInfoStdio(CFlag *ipFlag, const char *progname)
 #endif
 
   if (check_unicode_info(stdin, ipFlag, progname, &bomtype_orig))
-    RetVal = -1;
+    return -1;
 
-  /* info sucessful? */
+  /* info successful? */
 #ifdef D2U_UNICODE
-  if (!RetVal) {
-    if ((ipFlag->bomtype == FILE_UTF16LE) || (ipFlag->bomtype == FILE_UTF16BE)) {
-      FileInfoW(stdin, ipFlag, "", bomtype_orig, progname);
-    } else {
-      FileInfo(stdin, ipFlag, "", bomtype_orig, progname);
-    }
+  if ((ipFlag->bomtype == FILE_UTF16LE) || (ipFlag->bomtype == FILE_UTF16BE)) {
+    FileInfoW(stdin, ipFlag, "", bomtype_orig, progname);
+  } else {
+    FileInfo(stdin, ipFlag, "", bomtype_orig, progname);
   }
 #else
-  if (!RetVal)
-    FileInfo(stdin, ipFlag, "", bomtype_orig, progname);
+  FileInfo(stdin, ipFlag, "", bomtype_orig, progname);
 #endif
 
-  return RetVal;
+  return 0;
 }
 
 void get_info_options(char *option, CFlag *pFlag, const char *progname)
@@ -2220,6 +2221,9 @@ void get_info_options(char *option, CFlag *pFlag, const char *progname)
 
   while (*ptr != '\0') {
     switch (*ptr) {
+      case '0':   /* Print null characters instead of newline characters. */
+        pFlag->file_info |= INFO_PRINT0;
+        break;
       case 'd':   /* Print nr of DOS line breaks. */
         pFlag->file_info |= INFO_DOS;
         default_info = 0;
@@ -2277,7 +2281,6 @@ int parse_options(int argc, char *argv[],
   int ShouldExit = 0;
   int CanSwitchFileMode = 1;
   int process_options = 1;
-  int RetVal = 0;
 #ifdef D2U_UNIFILE
   char *ptr;
 #endif
@@ -2334,6 +2337,12 @@ int parse_options(int argc, char *argv[],
         pFlag->KeepDate = 1;
       else if ((strcmp(argv[ArgIdx],"-f") == 0) || (strcmp(argv[ArgIdx],"--force") == 0))
         pFlag->Force = 1;
+#ifndef NO_CHOWN
+      else if (strcmp(argv[ArgIdx],"--allow-chown") == 0)
+        pFlag->AllowChown = 1;
+      else if (strcmp(argv[ArgIdx],"--no-allow-chown") == 0)
+        pFlag->AllowChown = 0;
+#endif
 #ifdef D2U_UNICODE
 #if (defined(_WIN32) && !defined(__CYGWIN__))
       else if ((strcmp(argv[ArgIdx],"-gb") == 0) || (strcmp(argv[ArgIdx],"--gb18030") == 0))
@@ -2512,33 +2521,34 @@ int parse_options(int argc, char *argv[],
         pFlag->stdio_mode = 0;
       }
     } else {
-      pFlag->stdio_mode = 0;
       /* not an option */
+      int conversion_error;
+      pFlag->stdio_mode = 0;
       if (pFlag->NewFile) {
         if (CanSwitchFileMode)
           CanSwitchFileMode = 0;
         else {
 #ifdef D2U_UNICODE
-          RetVal = ConvertNewFile(argv[ArgIdx-1], argv[ArgIdx], pFlag, progname, Convert, ConvertW);
+          conversion_error = ConvertNewFile(argv[ArgIdx-1], argv[ArgIdx], pFlag, progname, Convert, ConvertW);
 #else
-          RetVal = ConvertNewFile(argv[ArgIdx-1], argv[ArgIdx], pFlag, progname, Convert);
+          conversion_error = ConvertNewFile(argv[ArgIdx-1], argv[ArgIdx], pFlag, progname, Convert);
 #endif
           if (pFlag->verbose)
-            print_messages_newfile(pFlag, argv[ArgIdx-1], argv[ArgIdx], progname, RetVal);
+            print_messages(pFlag, argv[ArgIdx-1], argv[ArgIdx], progname, conversion_error);
           CanSwitchFileMode = 1;
         }
       } else {
         if (pFlag->file_info) {
-          RetVal = GetFileInfo(argv[ArgIdx], pFlag, progname);
+          conversion_error = GetFileInfo(argv[ArgIdx], pFlag, progname);
           print_messages_info(pFlag, argv[ArgIdx], progname);
         } else {
 #ifdef D2U_UNICODE
-          RetVal = ConvertNewFile(argv[ArgIdx], argv[ArgIdx], pFlag, progname, Convert, ConvertW);
+          conversion_error = ConvertNewFile(argv[ArgIdx], argv[ArgIdx], pFlag, progname, Convert, ConvertW);
 #else
-          RetVal = ConvertNewFile(argv[ArgIdx], argv[ArgIdx], pFlag, progname, Convert);
+          conversion_error = ConvertNewFile(argv[ArgIdx], argv[ArgIdx], pFlag, progname, Convert);
 #endif
           if (pFlag->verbose)
-            print_messages_oldfile(pFlag, argv[ArgIdx], progname, RetVal);
+            print_messages(pFlag, argv[ArgIdx], NULL, progname, conversion_error);
         }
       }
     }
@@ -2720,7 +2730,7 @@ wint_t d2u_putwc(wint_t wc, FILE *f, CFlag *ipFlag, const char *progname)
        * Surrogate halves in UTF-8 are invalid. See also
        * http://en.wikipedia.org/wiki/UTF-8#Invalid_code_points
        * http://tools.ietf.org/html/rfc3629#page-5
-       * It is a bug in (some implemenatations of) wcstombs().
+       * It is a bug in (some implementations of) wcstombs().
        * On Cygwin 1.7 wcstombs() produces correct UTF-8 from UTF-16 surrogate pairs.
        */
       /* Decode UTF-16 surrogate pair */
